#!/usr/bin/env python3

import sys, os, subprocess, shutil

os.chdir (os.getenv ('srcdir', os.path.dirname (__file__)))

git = shutil.which ('git'); assert git
make = shutil.which ('make'); assert make
java = shutil.which ('java'); assert java
cxx = shutil.which ('c++'); assert cxx

pull = False
if not os.path.exists ('aots'):
	subprocess.run ([git, 'clone', 'https://github.com/adobe-type-tools/aots'], check=True)
	pull = True

if pull or 'pull' in sys.argv:
	subprocess.run ([git, 'pull'], cwd='aots', check=True)
	subprocess.run ([make, '-C', 'aots'], check=True)
	subprocess.run ([make, '-C', 'aots/harfbuzz'], check=True)

shutil.copy ('hb-aots-tester.cpp', 'aots/harfbuzz')
# TODO: remove *nix assumptions
subprocess.run ([cxx, '-std=c++11', '-Wno-narrowing', 'aots/harfbuzz/hb-aots-tester.cpp',
	'../../../../src/harfbuzz.cc', '-DHB_NO_MT', '-fno-exceptions', '-lm',
	'-I../../../../src', '-o', 'aots/harfbuzz/aots'], check=True)
shutil.rmtree ('tests')
os.mkdir ('tests')
subprocess.run (['./aots'], cwd='aots/harfbuzz', check=True)



#!/usr/bin/env python3

"""usage: ./gen-ucd-table [--rust] ucd.nounihan.grouped.xml [/path/to/hb-script-list.h]

Input file:
* https://unicode.org/Public/UCD/latest/ucdxml/ucd.nounihan.grouped.zip
"""

# https://github.com/harfbuzz/packtab
import packTab
import packTab.ucdxml

import sys, re
import logging

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.INFO)

if len(sys.argv) > 1 and sys.argv[1] == "--rust":
    del sys.argv[1]
    logging.info("Generating Rust code...")
    language = "rust"
else:
    logging.info("Generating C code...")
    language = "c"
language = packTab.languages[language]

if len(sys.argv) not in (2, 3):
    sys.exit(__doc__)

logging.info("Loading UCDXML...")
ucdxml = packTab.ucdxml.load_ucdxml(sys.argv[1])
ucd = packTab.ucdxml.ucdxml_get_repertoire(ucdxml)

hb_script_list_h = "hb-script-list.h" if len(sys.argv) < 3 else sys.argv[2]

logging.info("Preparing data tables...")


# This is how the data is encoded:
#
# General_Category (gc), Canonical_Combining_Class (ccc),
# and Script (sc) are encoded as integers.
#
# Mirroring character (bmg) is encoded as difference from
# the original character.
#
# Composition & Decomposition (dm) are encoded elaborately,
# as discussed below.

gc = [u["gc"] for u in ucd]
ccc = [int(u["ccc"]) for u in ucd]
bmg = [int(v, 16) - int(u) if v else 0 for u, v in enumerate(u["bmg"] for u in ucd)]
sc = [u["sc"] for u in ucd]


# Prepare Compose / Decompose data
#
# This code is very dense.  See hb_ucd_compose() / hb_ucd_decompose() for the logic.

dm = {
    i: tuple(int(v, 16) for v in u["dm"].split())
    for i, u in enumerate(ucd)
    if u["dm"] != "#" and u["dt"] == "can" and not (0xAC00 <= i < 0xAC00 + 11172)
}
ce = {i for i, u in enumerate(ucd) if u["Comp_Ex"] == "Y"}

assert not any(v for v in dm.values() if len(v) not in (1, 2))
dm1 = sorted(set(v for v in dm.values() if len(v) == 1))
assert all((v[0] >> 16) in (0, 2) for v in dm1)
dm1_p0_array = ["0x%04X" % (v[0] & 0xFFFF) for v in dm1 if (v[0] >> 16) == 0]
dm1_p2_array = ["0x%04X" % (v[0] & 0xFFFF) for v in dm1 if (v[0] >> 16) == 2]
dm1_order = {v: i + 1 for i, v in enumerate(dm1)}

dm2 = sorted(
    (v + (i if i not in ce and not ccc[i] else 0,), v)
    for i, v in dm.items()
    if len(v) == 2
)

filt = lambda v: (
    (v[0] & 0xFFFFF800) == 0x0000
    and (v[1] & 0xFFFFFF80) == 0x0300
    and (v[2] & 0xFFF0C000) == 0x0000
)
dm2_u32_array = [v for v in dm2 if filt(v[0])]
dm2_u64_array = [v for v in dm2 if not filt(v[0])]
assert dm2_u32_array + dm2_u64_array == dm2
dm2_u32_array = [
    "HB_CODEPOINT_ENCODE3_11_7_14 (0x%04X, 0x%04X, 0x%04X)" % v[0]
    for v in dm2_u32_array
]
dm2_u64_array = [
    "HB_CODEPOINT_ENCODE3 (0x%04X, 0x%04X, 0x%04X)" % v[0] for v in dm2_u64_array
]

l = 1 + len(dm1_p0_array) + len(dm1_p2_array)
dm2_order = {v[1]: i + l for i, v in enumerate(dm2)}

dm_order = {None: 0}
dm_order.update(dm1_order)
dm_order.update(dm2_order)


# Prepare General_Category / Script mapping arrays

gc_order = dict()
for i, v in enumerate(
    (
        "Cc",
        "Cf",
        "Cn",
        "Co",
        "Cs",
        "Ll",
        "Lm",
        "Lo",
        "Lt",
        "Lu",
        "Mc",
        "Me",
        "Mn",
        "Nd",
        "Nl",
        "No",
        "Pc",
        "Pd",
        "Pe",
        "Pf",
        "Pi",
        "Po",
        "Ps",
        "Sc",
        "Sk",
        "Sm",
        "So",
        "Zl",
        "Zp",
        "Zs",
    )
):
    gc_order[i] = v
    gc_order[v] = i

sc_order = dict()
sc_array = []
sc_re = re.compile(r"\b(HB_SCRIPT_[_A-Z]*).*HB_TAG [(]'(.)','(.)','(.)','(.)'[)]")
for line in open(hb_script_list_h):
    m = sc_re.search(line)
    if not m:
        continue
    name = m.group(1)
    tag = "".join(m.group(i) for i in range(2, 6))
    i = len(sc_array)
    sc_order[tag] = i
    sc_order[i] = tag
    if language.name == "rust":
        name = name.replace("HB_SCRIPT_", "script::")
    sc_array.append(name)


# Write out main data

DEFAULT = "DEFAULT"
COMPACT = "COMPACT"
SLOPPY = "SLOPPY"

compression_level = {
    DEFAULT: 3,
    COMPACT: 9,
    SLOPPY: 9,
}

logging.info("Generating output...")
print("/* == Start of generated table == */")
print("/*")
print(" * The following table is generated by running:")
print(" *")
print(
    " *   ./gen-ucd-table.py %sucd.nounihan.grouped.xml hb-script-list.h"
    % (("--%s " % language.name) if language.name != "c" else "")
)
print(" *")
print(" * on file with this description:", ucdxml.description)
print(" */")
print()
if language.name == "c":
    print("#ifndef HB_UCD_TABLE_HH")
    print("#define HB_UCD_TABLE_HH")
    print()
    print('#include "hb.hh"')
    print()
elif language.name == "rust":
    print("pub(crate) mod ucd {")
    print()
    print("#![allow(unused_parens)]")
    print("#![allow(clippy::unnecessary_cast, clippy::unreadable_literal, clippy::double_parens)]")
    print()
    print("use crate::hb::algs::{HB_CODEPOINT_ENCODE3, HB_CODEPOINT_ENCODE3_11_7_14};")
    print("use crate::hb::common::script;")
    print("use crate::hb::common::Script as hb_script_t;")
    print()
else:
    assert False, "Unknown language: %s" % language.name

# Write mapping data

uint16_t = language.type_name("u16")
uint32_t = language.type_name("u32")
uint64_t = language.type_name("u64")

if language.name == "c":
    private = True
elif language.name == "rust":
    private = False
else:
    assert False, "Unknown language: %s" % language.name

code = packTab.Code("_hb_ucd")
sc_array, _ = code.addArray("hb_script_t", "sc_map", sc_array)
dm1_p0_array, _ = code.addArray(uint16_t, "dm1_p0_map", dm1_p0_array)
dm1_p2_array, _ = code.addArray(uint16_t, "dm1_p2_map", dm1_p2_array)
dm2_u32_array, _ = code.addArray(uint32_t, "dm2_u32_map", dm2_u32_array)
dm2_u64_array, _ = code.addArray(uint64_t, "dm2_u64_map", dm2_u64_array)
code.print_code(language=language, private=private)

datasets = [
    ("gc", gc, "Cn", gc_order),
    ("ccc", ccc, 0, None),
    ("bmg", bmg, 0, None),
    ("sc", sc, "Zzzz", sc_order),
    ("dm", dm, None, dm_order),
]


# Write main data

modes = {}
if language.name == "c":
    modes[DEFAULT] = "#ifndef HB_OPTIMIZE_SIZE"
    modes[COMPACT] = "#elif !defined(HB_NO_UCD_UNASSIGNED)"
    modes[SLOPPY] = "#else"
    modes[None] = "#endif"
else:
    modes[DEFAULT] = ""

for step, text in modes.items():
    print()
    if text:
        print(text)
        print()
    if step is None:
        continue

    compression = compression_level[step]
    logging.info("  Compression=%d:" % compression)

    if step == SLOPPY:
        for i in range(len(gc)):
            if (i % 128) and gc[i] == "Cn":
                gc[i] = gc[i - 1]
        for i in range(len(gc) - 2, -1, -1):
            if ((i + 1) % 128) and gc[i] == "Cn":
                gc[i] = gc[i + 1]
        for i in range(len(sc)):
            if (i % 128) and sc[i] == "Zzzz":
                sc[i] = sc[i - 1]
        for i in range(len(sc) - 2, -1, -1):
            if ((i + 1) % 128) and sc[i] == "Zzzz":
                sc[i] = sc[i + 1]

    code = packTab.Code("_hb_ucd")

    for name, data, default, mapping in datasets:
        sol = packTab.pack_table(
            data, default, mapping=mapping, compression=compression
        )
        logging.info("      Dataset=%-8s FullCost=%d" % (name, sol.fullCost))
        sol.genCode(code, name, private=private, language=language)

    code.print_code(language=language)

    print()

if language.name == "c":
    print("#endif /* HB_UCD_TABLE_HH */")
elif language.name == "rust":
    print("}")
else:
    assert False, "Unknown language: %s" % language.name
print()
print("/* == End of generated table == */")
logging.info("Done.")



#!/usr/bin/env python3

import os
import re
import shlex
import subprocess
import sys
import tempfile
from difflib import unified_diff

print("TAP version 14")

args = sys.argv[1:]
if (
    not args
    or not os.path.exists(args[0])
    or "hb-vector" not in args[0]
):
    sys.exit("First argument does not seem to point to usable hb-vector.")
hb_vector, args = args[0], args[1:]

if not args:
    args = ["-"]

env = os.environ.copy()
env["LC_ALL"] = "C"
EXE_WRAPPER = os.environ.get("MESON_EXE_WRAPPER")


def resolve_path(base, path):
    if os.path.isabs(path):
        return path
    return os.path.normpath(os.path.join(base, path))


def open_vector_batch_process():
    cmd = [hb_vector, "--batch"]
    if EXE_WRAPPER:
        cmd = [EXE_WRAPPER] + cmd

    process = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        env=env,
        text=True,
    )
    return [process]


def close_vector_batch_process(vector_process):
    process = vector_process[0]
    if process is None:
        return

    if process.stdin:
        process.stdin.close()
    if process.stdout:
        process.stdout.close()

    process.wait()
    vector_process[0] = None


vector_process = open_vector_batch_process()


def build_vector_command(font_path, options, unicodes, output_path):
    command = []
    if options:
        command.extend(shlex.split(options))
    command.extend([f"--output-file={output_path}", f"--unicodes={unicodes}", font_path])
    return command


def run_vector_single(command):
    cmd = [hb_vector]
    cmd.extend(command)
    if EXE_WRAPPER:
        cmd = [EXE_WRAPPER] + cmd
    return subprocess.run(cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)


def vector_cmd(command, vector_process):
    process = vector_process[0]

    # (Re)start vector tool if it is dead.
    if process.poll() is not None:
        if process.stdin:
            process.stdin.close()
        if process.stdout:
            process.stdout.close()
        vector_process[0] = open_vector_batch_process()[0]
        process = vector_process[0]

    process.stdin.write(";".join(command) + "\n")
    process.stdin.flush()

    diagnostics = []
    while True:
        line = process.stdout.readline()
        if line == "":
            return None, diagnostics

        line = line.rstrip("\n")
        if line in ("success", "failure"):
            return line, diagnostics

        diagnostics.append(line)


def run_vector(font_path, options, unicodes, output_path):
    command = build_vector_command(font_path, options, unicodes, output_path)
    status, diagnostics = vector_cmd(command, vector_process)

    if status == "success":
        return subprocess.CompletedProcess(
            args=command,
            returncode=0,
            stdout="",
            stderr="\n".join(diagnostics),
        )

    # Fall back to one-shot execution for proper diagnostics.
    result = run_vector_single(command)
    if diagnostics:
        stderr = "\n".join(diagnostics)
        if result.stderr:
            stderr += "\n" + result.stderr
        result = subprocess.CompletedProcess(
            args=result.args,
            returncode=result.returncode,
            stdout=result.stdout,
            stderr=stderr,
        )
    return result


def parse_unicodes(s):
    # Match test/shape/hb_test_tools.py Unicode.parse semantics.
    s = re.sub(r"0[xX]", " ", s)
    s = re.sub(r"[<+\->{},;&#\\xXuUnNiI\n\t]", " ", s)
    try:
        return "".join(chr(int(x, 16)) for x in s.split())
    except ValueError as e:
        raise ValueError(f"invalid unicode codepoint list: {s!r}") from e


number = 0
fails = 0

try:
    for filename in args:
        if filename == "-":
            f = sys.stdin
            base = os.getcwd()
            print("# Running tests from standard input")
        else:
            f = open(filename, encoding="utf-8")
            base = os.path.dirname(os.path.abspath(filename))
            print("# Running tests in " + filename)

        for lineno, line in enumerate(f, start=1):
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            number += 1
            fields = line.split(";", 3)
            if len(fields) != 4:
                fails += 1
                print(f"not ok {number} - parse:{filename}:{lineno}")
                print("# malformed test line, expected 4 ';'-separated fields")
                continue

            font_file, options, unicodes, expected_file = fields
            name = expected_file
            font_path = resolve_path(base, font_file)
            expected_path = resolve_path(base, expected_file)

            if not os.path.exists(font_path):
                fails += 1
                print(f"not ok {number} - {name}")
                print(f"# font file not found: {font_path}")
                continue

            if not os.path.exists(expected_path):
                fails += 1
                print(f"not ok {number} - {name}")
                print(f"# expected output file not found: {expected_path}")
                continue

            try:
                parse_unicodes(unicodes)
            except ValueError as e:
                fails += 1
                print(f"not ok {number} - {name}")
                print("# " + str(e))
                continue

            with tempfile.NamedTemporaryFile(prefix="hb-vector-test-", suffix=".svg", delete=False) as tmp:
                output_path = tmp.name

            try:
                result = run_vector(font_path, options, unicodes, output_path)
                if result.returncode != 0:
                    fails += 1
                    print(f"not ok {number} - {name}")
                    print(f"# hb-vector exited with code {result.returncode}")
                    stderr = result.stderr.strip()
                    if stderr:
                        for l in stderr.splitlines():
                            print("# " + l)
                    continue

                if not os.path.exists(output_path):
                    fails += 1
                    print(f"not ok {number} - {name}")
                    print("# hb-vector did not create output file")
                    continue

                with open(expected_path, encoding="utf-8") as fp:
                    expected = fp.read()
                with open(output_path, encoding="utf-8") as fp:
                    actual = fp.read()

                if actual == expected:
                    print(f"ok {number} - {name}")
                    continue

                fails += 1
                print(f"not ok {number} - {name}")
                diff = unified_diff(
                    expected.splitlines(keepends=True),
                    actual.splitlines(keepends=True),
                    fromfile=expected_path,
                    tofile="actual",
                )
                for diff_line in diff:
                    print("# " + diff_line.rstrip("\n"))
            finally:
                if os.path.exists(output_path):
                    os.unlink(output_path)

        if f is not sys.stdin:
            f.close()
finally:
    close_vector_batch_process(vector_process)

print(f"1..{number}")

if fails:
    sys.exit(1)



#!/usr/bin/env python3

"""usage: ./gen-arabic-pua.py
"""

import packTab


print ("/* == Start of generated table == */")
print ("/*")
print (" * The following table is generated by running:")
print (" *")
print (" *   ./gen-arabic-pua.py")
print (" *")
print (" */")
print ()
print ("#ifndef HB_OT_SHAPER_ARABIC_PUA_HH")
print ("#define HB_OT_SHAPER_ARABIC_PUA_HH")
print ()
print ('#include "hb.hh"')
print ()

code = packTab.Code('_hb_arabic_pua')

for p in ("ArabicPUASimplified.txt", "ArabicPUATraditional.txt"):
    with open (p, encoding='utf-8') as f:
        fields = [l.split('\t') for l in f if l[:1] != '#']
    data = {int(fs[1], 16):int(fs[0], 16) for fs in fields}
    sol = packTab.pack_table(data, compression=9)
    sol.genCode(code, f'{p[9:13].lower()}_map')

code.print_c()

print ()
print ("#endif /* HB_OT_SHAPER_ARABIC_PUA_HH */")
print ()
print ("/* == End of generated table == */")



#!/usr/bin/env python3

import sys, os, subprocess, shutil, glob
import xml.etree.ElementTree as ET

# Can we extract this from HTML element itself? I couldn't.
namespaces = {
    "ft": "https://github.com/OpenType/fonttest",
    "xlink": "http://www.w3.org/1999/xlink",
}


def ns(s):
    ns, s = s.split(":")
    return "{%s}%s" % (namespaces[ns], s)


def unistr(s):
    return ",".join("U+%04X" % ord(c) for c in s)


def glyphstr(glyphs):
    out = []
    for glyphname, x, y in glyphs:
        if x or y:
            out.append("%s@%d,%d" % (glyphname, x, y))
        else:
            out.append(glyphname)
    return "[" + "|".join(out) + "]"


def extract_tests(input):
    html = ET.fromstring(input)
    found = False

    result = []

    for elt in html.findall(".//*[@class='expected'][@ft:id]", namespaces):
        found = True
        name = elt.get(ns("ft:id"))
        text = elt.get(ns("ft:render"))
        font = elt.get(ns("ft:font"))
        variations = elt.get(ns("ft:var"), "").replace(":", "=").replace(";", ",")
        glyphs = []
        for use in elt.findall(".//use"):
            x = int(use.get("x"))
            y = int(use.get("y"))
            href = use.get(ns("xlink:href"))
            assert href[0] == "#"
            glyphname = ".".join(href[1:].split("/")[1].split(".")[1:])
            glyphs.append((glyphname, x, y))
        opts = "--font-size=1000 --ned --remove-default-ignorables"
        if variations:
            opts = opts + " --variations=%s" % variations
        result.append(
            "../fonts/%s;%s;%s;%s" % (font, opts, unistr(text), glyphstr(glyphs))
        )

    for elt in html.findall(".//*[@class='expected-no-crash'][@ft:id]", namespaces):
        found = True
        name = elt.get(ns("ft:id"))
        text = elt.get(ns("ft:render"))
        font = elt.get(ns("ft:font"))
        variations = elt.get(ns("ft:var"), "").replace(":", "=").replace(";", ",")
        opts = ""
        if variations:
            opts = "--variations=%s" % variations
        result.append("../fonts/%s;%s;%s;*" % (font, opts, unistr(text)))

    assert found
    return "\n".join(result) + "\n"


os.chdir(os.getenv("srcdir", os.path.dirname(__file__)))

git = shutil.which("git")
assert git

if os.path.isdir("./text-rendering-tests"):
    subprocess.run([git, "pull"], cwd="text-rendering-tests", check=True)
else:
    subprocess.run(
        [git, "clone", "https://github.com/unicode-org/text-rendering-tests"],
        check=True,
    )

shutil.rmtree("fonts", ignore_errors=True)
assert not os.path.exists("fonts")
shutil.copytree("text-rendering-tests/fonts", "fonts")
subprocess.run([git, "add", "fonts"], check=True)

shutil.rmtree("tests", ignore_errors=True)
assert not os.path.isdir("tests")
os.mkdir("tests")

with open("DISABLED", "r") as f:
    disabled = f.read()

tests = []
disabled_tests = []

for x in sorted(os.listdir("text-rendering-tests/testcases")):
    if not x.endswith(".html") or x == "index.html":
        continue
    out = "tests/%s.tests" % x.split(".html")[0]
    with open("text-rendering-tests/testcases/" + x, "r") as f:
        content = f.read()
    with open(out, "w") as f:
        f.write("@font-funcs=ot,ft\n")
        f.write(extract_tests(content))
    if out in disabled:
        disabled_tests.append(out)
    else:
        tests.append(out)

subprocess.run([git, "add", "tests"], check=True)

with open("meson.build", "w") as f:
    f.write(
        "\n".join(
            ["text_rendering_tests_tests = ["]
            + ["  '%s'," % x.split("tests/")[1] for x in tests]
            + ["]", "", "disabled_text_rendering_tests_tests = ["]
            + ["  '%s'," % x.split("tests/")[1] for x in disabled_tests]
            + ["]", ""]
        )
    )

subprocess.run([git, "add", "meson.build"], check=True)

print(
    'Updated the testsuit, now run `git commit -e -m "[test/text-rendering-tests] Update from upstream"`'
)



#!/usr/bin/env python3

import os
import sys

srcdir = sys.argv[1]
base_srcdir = sys.argv[2]
builddir = sys.argv[3]

os.chdir(srcdir)


def removeprefix(s):
    abs_path = os.path.join(base_srcdir, s)
    return os.path.relpath(abs_path, srcdir)


HBHEADERS = [os.path.basename(x) for x in os.getenv("HBHEADERS", "").split()] or [
    x for x in os.listdir(".") if x.startswith("hb") and x.endswith(".h")
]
HBHEADERS = [x for x in HBHEADERS if x.endswith(".h")]
HBSOURCES = [removeprefix(x) for x in os.getenv("HBSOURCES", "").split()] or [
    x for x in os.listdir(".") if x.startswith("hb") and x.endswith((".cc", ".hh"))
]
stat = 0

for x in HBHEADERS:
    print(f"Checking {x}")
    with open(x, "r", encoding="utf-8") as f:
        content = f.read()
    if ("HB_BEGIN_DECLS" not in content) or ("HB_END_DECLS" not in content):
        print(
            "Ouch, file %s does not have HB_BEGIN_DECLS / HB_END_DECLS, but it should"
            % x
        )
        stat = 1

for x in HBSOURCES:
    print(f"Checking {x}")
    with open(x, "r", encoding="utf-8") as f:
        content = f.read()
    if ("HB_BEGIN_DECLS" in content) or ("HB_END_DECLS" in content):
        print("Ouch, file %s has HB_BEGIN_DECLS / HB_END_DECLS, but it shouldn't" % x)
        stat = 1

sys.exit(stat)



#!/usr/bin/env python3

import sys, os, subprocess, hashlib

print("TAP version 14")

args = sys.argv[1:]

verbose = False
if args and args[0] == "-v":
    verbose = True
    args = args[1:]

if not args or args[0].find("hb-shape") == -1 or not os.path.exists(args[0]):
    sys.exit("""First argument does not seem to point to usable hb-shape.""")
hb_shape, args = args[0], args[1:]

env = os.environ.copy()
env["LC_ALL"] = "C"

EXE_WRAPPER = os.environ.get("MESON_EXE_WRAPPER")


def open_shape_batch_process():
    cmd = [hb_shape, "--batch"]
    if EXE_WRAPPER:
        cmd = [EXE_WRAPPER] + cmd

    process = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
    )
    return [process]


shape_process = open_shape_batch_process()
no_glyph_names_process = None


def shape_cmd(command, shape_process, verbose=False):
    global hb_shape

    # (Re)start shaper if it is dead
    if shape_process[0].poll() is not None:
        shape_process[0].stdin.close()
        shape_process[0].stdout.close()
        shape_process[0].stderr.close()
        shape_process[0] = open_shape_batch_process()[0]

    if verbose:
        print("# " + hb_shape + " " + " ".join(command))
    shape_process[0].stdin.write((";".join(command) + "\n").encode("utf-8"))
    shape_process[0].stdin.flush()
    return shape_process[0].stdout.readline().decode("utf-8").strip()


def plural(what):
    if not what.endswith("s"):
        what += "s"
    return what


def whats_var_name(what):
    return plural(what).replace("-", "_")


def supported_whats_var_name(what):
    whats = whats_var_name(what)
    return "supported_" + whats


def supported_whats(what):
    return globals()[supported_whats_var_name(what)]


def all_whats_var_name(what):
    whats = whats_var_name(what)
    return "all_" + whats


def all_whats(what):
    return globals()[all_whats_var_name(what)]


# Collect supported backends
for what in ["shaper", "face-loader", "font-funcs"]:
    subcommand = "--list-" + plural(what)

    cmd = [hb_shape, subcommand]
    if EXE_WRAPPER:
        cmd = [EXE_WRAPPER] + cmd

    what_process = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=sys.stdout,
        env=env,
    )
    # Capture the output
    what_list = what_process.communicate()[0].decode("utf-8").strip().split()
    if what_process.returncode:
        sys.exit(f"Failed to run: {hb_shape} {subcommand}")
    whats = plural(what)
    var_name = supported_whats_var_name(what)
    globals()[var_name] = what_list
    print(f"# Supported {whats}: {what_list}")

# If running under Wine and not native dlls, make the respective shapers unavailable.
if os.environ.get("WINEPATH"):
    overrides = os.environ.get("WINEDLLOVERRIDES", "").lower()
    if "directwrite" in supported_shapers and overrides.find("dwrite") == -1:
        supported_shapers.remove("directwrite")
        print("# Skipping DirectWrite shaper under Wine.")
    if "uniscribe" in supported_shapers and overrides.find("usp10") == -1:
        supported_shapers.remove("uniscribe")
        print("# Skipping Uniscribe shaper under Wine.")


number = 0
passes = 0
fails = 0
skips = 0

if not len(args):
    args = ["-"]

for filename in args:
    if filename == "-":
        print("# Running tests from standard input")
    else:
        print("# Running tests in " + filename)

    if filename == "-":
        f = sys.stdin
    else:
        f = open(filename, encoding="utf8")

    # By default test all backends
    for what in ["shaper", "face-loader", "font-funcs"]:
        all_var_name = all_whats_var_name(what)
        globals()[all_var_name] = supported_whats(what)
    all_face_loaders = os.environ.get("HB_FACE_LOADER", "ot").split(",")
    all_shapers = os.environ.get("HB_SHAPER_LIST", "ot").split(",")

    # Right now we only test the 'ot' shaper if nothing specified,
    # but try all font-funcs unless overriden.
    # Only 'ot' face-loader is tested.

    for line in f:
        comment = False
        if line.startswith("#"):
            comment = True
            line = line[1:]

            if line.startswith(" "):
                if verbose:
                    print("#%s" % line)
                continue

        line = line.strip()
        if not line:
            continue

        if line.startswith("@"):
            # Directive
            line = line.strip()
            line = line.split("#")[0].strip()[1:]
            consumed = False
            for what in ["shaper", "face-loader", "font-funcs"]:
                whats = plural(what)
                if line.startswith(what) or line.startswith(whats):
                    command, values = line.split("=")
                    values = values.strip().split(",")

                    supported = supported_whats(what)
                    if command[-1] == "-":
                        # Exclude
                        values = [v for v in supported if v not in values]
                    else:
                        # Specify
                        values = [v for v in values if v in supported]

                    var_name = all_whats_var_name(what)
                    print(f"# Setting {whats} to test to {values}")
                    globals()[var_name] = values
                    consumed = True
            if consumed:
                print("#", line)
                continue
            else:
                print("Unrecognized directive: %s" % line, file=sys.stderr)
                sys.exit(1)

        fontfile, options, unicodes, glyphs_expected = line.split(";")
        options = options.split()
        if fontfile.startswith("/") or fontfile.startswith('"/'):
            if os.name == "nt":  # Skip on Windows
                continue

            fontfile, expected_hash = (fontfile.split("@") + [""])[:2]

            try:
                with open(fontfile, "rb") as ff:
                    if expected_hash:
                        actual_hash = hashlib.sha1(ff.read()).hexdigest().strip()
                        if actual_hash != expected_hash:
                            skips += 1
                            print(
                                "# Different version of %s found; Expected hash %s, got %s; skipping."
                                % (fontfile, expected_hash, actual_hash)
                            )
                            continue
            except IOError:
                skips += 1
                print("# %s not found, skip." % fontfile)
                continue
        else:
            cwd = os.path.dirname(filename)
            fontfile = os.path.normpath(os.path.join(cwd, fontfile))

        if comment:
            if verbose:
                print('# # %s "%s" --unicodes %s' % (hb_shape, fontfile, unicodes))
            continue

        skip_test = False
        shaper = None
        face_loader = None
        font_funcs = None
        new_options = []
        it = iter(options)
        for option in it:
            consumed = False
            for what in ["shaper", "face-loader", "font-funcs"]:
                if option.startswith("--" + what):
                    try:
                        backend = option.split("=")[1]
                    except IndexError:
                        backend = next(it)
                    if backend not in supported_whats(what):
                        skips += 1
                        number += 1
                        print(
                            f"ok {number} - {fontfile} # skip {what}={backend} not supported"
                        )
                        print(f"# Skipping test with {what}={backend}.")
                        skip_test = True
                        break
                    what = what.replace("-", "_")
                    globals()[what] = backend
                    consumed = True
            if not consumed:
                new_options.append(option)

            if skip_test:
                break
        if skip_test:
            continue
        options = new_options

        for face_loader in [face_loader] if face_loader else all_whats("face-loader"):
            for shaper in [shaper] if shaper else all_whats("shaper"):
                all_font_funcs_for_shaper = (
                    all_whats("font-funcs") if shaper == "ot" else [None]
                )
                for font_funcs in (
                    [font_funcs] if font_funcs else all_font_funcs_for_shaper
                ):
                    number += 1
                    extra_options = []

                    if shaper:
                        extra_options.append("--shaper=" + shaper)
                    if face_loader:
                        extra_options.append("--face-loader=" + face_loader)
                    if font_funcs:
                        extra_options.append("--font-funcs=" + font_funcs)

                    if glyphs_expected != "*":
                        extra_options.append("--verify")
                        extra_options.append("--unsafe-to-concat")
                        extra_options.append("--safe-to-insert-tatweel")

                    if verbose:
                        print(
                            "# shaper=%s face-loader=%s font-funcs=%s"
                            % (shaper, face_loader, font_funcs)
                        )
                    cmd = (
                        [fontfile] + ["--unicodes", unicodes] + options + extra_options
                    )
                    glyphs = shape_cmd(cmd, shape_process, verbose).strip()

                    if glyphs_expected == "*":
                        passes += 1
                        print(f"ok {number} - {fontfile}")
                        continue

                    final_glyphs = glyphs
                    final_glyphs_expected = glyphs_expected

                    if glyphs != glyphs_expected and glyphs.find("gid") != -1:
                        if not no_glyph_names_process:
                            no_glyph_names_process = open_shape_batch_process()

                        cmd2 = [fontfile] + ["--glyphs", "--no-glyph-names", glyphs]
                        final_glyphs = shape_cmd(cmd2, no_glyph_names_process).strip()

                        cmd2 = [fontfile] + [
                            "--glyphs",
                            "--no-glyph-names",
                            glyphs_expected,
                        ]
                        final_glyphs_expected = shape_cmd(
                            cmd2, no_glyph_names_process
                        ).strip()

                    # If the removal of glyph_ids failed, fail the test.
                    # https://github.com/harfbuzz/harfbuzz/issues/5169
                    if (
                        not final_glyphs_expected
                        or final_glyphs != final_glyphs_expected
                    ):
                        fails += 1
                        cmd = hb_shape + " " + " ".join(cmd)
                        print(f"not ok {number} - {cmd}")
                        print("   ---", file=sys.stderr)
                        print('   test_file: "' + filename + '"', file=sys.stderr)
                        print('   cmd: "' + cmd + '"', file=sys.stderr)
                        print('   actual:   "' + glyphs + '"', file=sys.stderr)
                        print('   expected: "' + glyphs_expected + '"', file=sys.stderr)
                        if final_glyphs != glyphs:
                            print(
                                '   actual_gids:   "' + final_glyphs + '"',
                                file=sys.stderr,
                            )
                            print(
                                '   expected_gids: "' + final_glyphs_expected + '"',
                                file=sys.stderr,
                            )
                        print("   ...", file=sys.stderr)
                    else:
                        cmd = hb_shape + " " + " ".join(cmd)
                        passes += 1
                        print(f"ok {number} - {cmd}")

print("# %d tests passed; %d failed; %d skipped." % (passes, fails, skips))
if not (fails + passes):
    print("# No tests ran.")
elif not (fails + skips):
    print("# All tests passed.")

print("1..%d" % number)



#!/usr/bin/env python3

"This tool is intended to be used from meson"

import os, sys, shutil, re

if len (sys.argv) < 4:
	sys.exit(__doc__)

version = sys.argv[1]
major, minor, micro = version.split (".")

OUTPUT = sys.argv[2]
INPUT = sys.argv[3]
CURRENT_SOURCE_DIR = os.path.dirname(INPUT)

try:
	with open (OUTPUT, "r", encoding='utf-8') as old_output:
		for line in old_output:
			old_version = re.match (r"#define HB_VERSION_STRING \"(\d.\d.\d)\"", line)
			if old_version and old_version[1] == version:
				sys.exit ()
except IOError:
	pass

with open (INPUT, "r", encoding='utf-8') as template:
	with open (OUTPUT, "wb") as output:
		output.write (template.read ()
			.replace ("@HB_VERSION_MAJOR@", major)
			.replace ("@HB_VERSION_MINOR@", minor)
			.replace ("@HB_VERSION_MICRO@", micro)
			.replace ("@HB_VERSION@", version)
			.encode ())

# copy it also to the source tree, but only if it has changed
baseline_filename = os.path.join (CURRENT_SOURCE_DIR, os.path.basename (OUTPUT))
with open(baseline_filename, "rb") as baseline:
	with open(OUTPUT, "rb") as generated:
		if baseline.read() != generated.read():
			shutil.copyfile (OUTPUT, baseline_filename)



#!/usr/bin/env python3

"""usage: ./gen-emoji-table.py [--rust] emoji-data.txt emoji-test.txt

Input file:
* https://www.unicode.org/Public/UCD/latest/ucd/emoji/emoji-data.txt
* https://www.unicode.org/Public/emoji/latest/emoji-test.txt
"""

import os
import sys
from collections import OrderedDict
import packTab

if len(sys.argv) > 1 and sys.argv[1] == "--rust":
	del sys.argv[1]
	language = packTab.languages["rust"]
else:
	language = packTab.languages["c"]

if len (sys.argv) != 3:
	sys.exit (__doc__)

f = open(sys.argv[1])
header = [f.readline () for _ in range(10)]

ranges = OrderedDict()
for line in f.readlines():
	line = line.strip()
	if not line or line[0] == '#':
		continue
	rang, typ = [s.strip() for s in line.split('#')[0].split(';')[:2]]

	rang = [int(s, 16) for s in rang.split('..')]
	if len(rang) > 1:
		start, end = rang
	else:
		start = end = rang[0]

	if typ not in ranges:
		ranges[typ] = []
	if ranges[typ] and ranges[typ][-1][1] == start - 1:
		ranges[typ][-1] = (ranges[typ][-1][0], end)
	else:
		ranges[typ].append((start, end))



print ("/* == Start of generated table == */")
print ("/*")
print (" * The following tables are generated by running:")
print (" *")
print (" *   ./gen-emoji-table.py %semoji-data.txt" %
		("--rust " if language.name == "rust" else ""))
print (" *")
print (" * on file with this header:")
print (" *")
for l in header:
	print (" * %s" % (l.strip()))
print (" */")
print ()
if language.name == "c":
	print ("#ifndef HB_UNICODE_EMOJI_TABLE_HH")
	print ("#define HB_UNICODE_EMOJI_TABLE_HH")
	print ()
	print ('#include "hb-unicode.hh"')
	print ()
elif language.name == "rust":
	print ("#![allow(unused_parens)]")
	print ("#![allow(clippy::unnecessary_cast, clippy::unreadable_literal, clippy::double_parens)]")
	print ()
	print ("use crate::hb::unicode::Codepoint;")
	print ()
else:
	assert False, "Unknown language: %s" % language.name

for typ, s in ranges.items():
	if typ != "Extended_Pictographic": continue

	arr = dict()
	for start,end in s:
		for i in range(start, end + 1):
			arr[i] = 1

	sol = packTab.pack_table(arr, 0, compression=9)
	code = packTab.Code('_hb_emoji')
	if language.name == "c":
		sol.genCode(code, 'is_'+typ, language=language)
		code.print_code(language=language)
		print()
	elif language.name == "rust":
		sol.genCode(code, 'is_'+typ+'_u8', language=language, private=False)
		code.print_code(language=language, private=False)
		print()
		print ("#[inline]")
		print ("pub(crate) fn is_%s (u: Codepoint) -> bool" % typ)
		print ("{")
		print ("  _hb_emoji_is_%s_u8 (u as usize) != 0" % typ)
		print ("}")
		print()
	else:
		assert False, "Unknown language: %s" % language.name

print ()
if language.name == "c":
	print ("#endif /* HB_UNICODE_EMOJI_TABLE_HH */")
print ()
print ("/* == End of generated table == */")


# Generate test file.
sequences = []
with open(sys.argv[2]) as f:
    for line in f.readlines():
        if "#" in line:
            line = line[:line.index("#")]
        if ";" in line:
            line = line[:line.index(";")]
        line = line.strip()
        line = line.split(" ")
        if len(line) < 2:
            continue
        sequences.append(line)

test_path = os.path.join(os.path.dirname(__file__), "..", "test", "shape", "data", "in-house", "tests", "emoji-clusters.tests")
if os.path.isdir(os.path.dirname(test_path)):
    try:
        with open(test_path, "w") as f:
            for sequence in sequences:
                f.write("../fonts/AdobeBlank2.ttf;--no-glyph-names --no-positions --font-funcs=ot")
                f.write(";" + ",".join(sequence))
                f.write(";[" + "|".join("1=0" for c in sequence) + "]\n")
    except OSError:
        pass



#!/usr/bin/env python3

import sys
import os

if len(sys.argv) < 2:
    print("Usage: python glob.py [DIR]...")
    sys.exit(1)

for arg in sys.argv[1:]:
    if os.path.isdir(arg):
        for dirpath, dirnames, filenames in os.walk(arg):
            for filename in filenames:
                print(os.path.join(dirpath, filename))
    elif os.path.isfile(arg):
        print(arg)



#!/usr/bin/env python3

"""Generates the code for a sorted unicode range array as used in hb-ot-os2-unicode-ranges.hh
Input is a tab separated list of unicode ranges from the otspec
(https://docs.microsoft.com/en-us/typography/opentype/spec/os2#ur).
"""

import re
import sys


print ("""static OS2Range _hb_os2_unicode_ranges[] =
{""")

args = sys.argv[1:]
input_file = args[0]

with open (input_file, mode="r", encoding="utf-8") as f:

  all_ranges = []
  current_bit = 0
  while True:
    line = f.readline().strip()
    if not line:
      break
    fields = re.split(r'\t+', line)
    if len(fields) == 3:
      current_bit = fields[0]
      fields = fields[1:]
    elif len(fields) > 3:
      raise Exception("bad input :(.")

    name = fields[0]
    ranges = re.split("-", fields[1])
    if len(ranges) != 2:
      raise Exception("bad input :(.")

    v = tuple((int(ranges[0], 16), int(ranges[1], 16), int(current_bit), name))
    all_ranges.append(v)

all_ranges = sorted(all_ranges, key=lambda t: t[0])

for ranges in all_ranges:
  start = ("0x%X" % ranges[0]).rjust(8)
  end = ("0x%X" % ranges[1]).rjust(8)
  bit = ("%s" % ranges[2]).rjust(3)

  print ("  {%s, %s, %s}, // %s" % (start, end, bit, ranges[3]))

print ("""};""")



#!/usr/bin/env python3
import sys


news_path = sys.argv[1]
release = sys.argv[2]

with open(news_path, "r", encoding="utf-8") as news_file:
    lines = news_file.readlines()

start = None
end = None
for i, line in enumerate(lines):
    line = line.rstrip()
    if line.startswith("Overview of changes leading to"):
        if start is not None:  # Start of next release
            end = i
            break
        if line.endswith(release):  # Start of the release
            start = i + 3  # Skip the header lines

assert start and end and end > start

print("".join(lines[start:end]))



#!/usr/bin/env python3

import sys
from fontTools.ttLib import TTFont
from fontTools.ttLib.tables.DefaultTable import DefaultTable

if len(sys.argv) == 1:
    print("usage: python addTable.py input.ttf output.ttf Wasm.bin")
    sys.exit(1)

font = TTFont(sys.argv[1])

wasm_table = DefaultTable("Wasm")
wasm_table.data = open(sys.argv[3], "rb").read()

font["Wasm"] = wasm_table

font.save(sys.argv[2])



#!/usr/bin/env python3

import os
import sys

from hb_fuzzing_tools import (
    fail,
    format_leading_comment,
    main_with_user_errors,
    parse_output_arg,
    print_recorded_seed_path,
    write_seed_file,
)


COMMAND_NAMES = ("hb-shape", "hb-raster", "hb-vector")


def usage(prog: str) -> str:
    return f"""Usage: {prog} [-o OUTPUT_OR_DIR] HB_{'{SHAPE|RASTER|VECTOR}'}_COMMAND FONT_FILE [HB_TOOL_ARGS...]

Generate a shared text/render seed using the hb-shape-style payload format:
an optional leading comment line containing the exact CLI invocation,
followed by the raw font bytes.

This recorder does not encode text, features, or variations into the payload.
Those are preserved only in the leading comment line for human recovery.

Options:
  -o OUTPUT_OR_DIR
              Write the seed to this file, or into this directory using
              <sha1(font)>.seed.
              Default: test/fuzzing/fonts/<sha1(font)>.seed
  -h, --help  Show this help.
"""


def find_font_file(args: list[str]) -> str:
    if len(args) < 2:
        fail("Expected at least a HarfBuzz text/render command and FONT_FILE.")

    command = args[0]
    base = os.path.basename(command)
    if not any(name in base for name in COMMAND_NAMES):
        fail(f"First argument does not look like one of {', '.join(COMMAND_NAMES)}: '{command}'.")

    font_file = args[1]
    if font_file.startswith("-"):
        fail("Expected FONT_FILE as the second positional argument.")

    return font_file


def record_test_main(argv: list[str]) -> int:
    usage_text = usage(argv[0])
    out, args = parse_output_arg(argv, usage_text)

    if len(args) < 2:
        print(usage_text, end="", file=sys.stderr)
        return 1

    font_file = find_font_file(args)
    seed_path = write_seed_file(
        font_file,
        out,
        bytearray(),
        default_suffix=".seed",
        leading_comment=format_leading_comment(args),
    )
    print_recorded_seed_path(seed_path)
    return 0


if __name__ == "__main__":
    raise SystemExit(main_with_user_errors(record_test_main, sys.argv))



# harfbuzz-icu links to libstdc++ because icu does.
CHECK_LIBSTDCXX_LIBS = [
    "harfbuzz",
    "harfbuzz-subset",
    "harfbuzz-raster",
    "harfbuzz-vector",
    "harfbuzz-gobject",
    "harfbuzz-cairo",
]

CHECK_SYMBOL_LIBS = CHECK_LIBSTDCXX_LIBS + [
    "harfbuzz-icu",
]



#!/usr/bin/env python3

"usage: gen-def.py harfbuzz.def hb.h [hb-blob.h hb-buffer.h ...]"

import os, re, sys

if len (sys.argv) < 3:
	sys.exit(__doc__)

output_file = sys.argv[1]
header_paths = sys.argv[2:]

headers_content = []
for h in header_paths:
	if h.endswith (".h"):
		with open (h, encoding='utf-8') as f: headers_content.append (f.read ())

symbols = sorted (re.findall (r"^hb_\w+(?= \()", "\n".join (headers_content), re.M))
if '--experimental-api' not in sys.argv:
	# Move these to harfbuzz-sections.txt when got stable
	experimental_symbols = \
"""hb_shape_justify
hb_subset_input_override_name_table
hb_subset_cff_get_charstring_data
hb_subset_cff_get_charstrings_index
hb_subset_cff2_get_charstring_data
hb_subset_cff2_get_charstrings_index
""".splitlines ()
	symbols = [x for x in symbols if x not in experimental_symbols]
symbols = "\n".join (symbols)

result = symbols if os.getenv ('PLAIN_LIST', '') else """EXPORTS
%s
LIBRARY lib%s-0.dll""" % (symbols, output_file.replace ('src/', '').replace ('.def', ''))

with open (output_file, "w") as f: f.write (result)



#!/usr/bin/env python3

import hashlib
import math
import os
import shlex
import struct
import sys


HB_FUZZING_EXTENDED_MAGIC = b"HBSUBFZ2"

HB_SUBSET_SETS_GLYPH_INDEX = 0
HB_SUBSET_SETS_UNICODE = 1
HB_SUBSET_SETS_DROP_TABLE_TAG = 3
HB_SUBSET_SETS_NAME_ID = 4
HB_SUBSET_SETS_NAME_LANG_ID = 5
HB_SUBSET_SETS_LAYOUT_FEATURE_TAG = 6
HB_SUBSET_SETS_LAYOUT_SCRIPT_TAG = 7

HB_FUZZING_OP_SET_FLAGS = 1
HB_FUZZING_OP_KEEP_EVERYTHING = 2
HB_FUZZING_OP_SET_CLEAR = 3
HB_FUZZING_OP_SET_INVERT = 4
HB_FUZZING_OP_SET_ADD_RANGES = 5
HB_FUZZING_OP_SET_DEL_RANGES = 6
HB_FUZZING_OP_TEXT_ADD = 7
HB_FUZZING_OP_TEXT_DEL = 8
HB_FUZZING_OP_AXIS_PIN_ALL_TO_DEFAULT = 9
HB_FUZZING_OP_AXIS_SET = 10

HB_FUZZING_AXIS_PIN_TO_DEFAULT = 0
HB_FUZZING_AXIS_SET_RANGE = 1

FLAG_BITS = {
    "--no-hinting": 0x00000001,
    "--retain-gids": 0x00000002,
    "--desubroutinize": 0x00000004,
    "--name-legacy": 0x00000008,
    "--set-overlaps-flag": 0x00000010,
    "--passthrough-tables": 0x00000020,
    "--notdef-outline": 0x00000040,
    "--glyph-names": 0x00000080,
    "--no-prune-unicode-ranges": 0x00000100,
    "--no-layout-closure": 0x00000200,
    "--optimize": 0x00000400,
    "--no-bidi-closure": 0x00000800,
    "--iftb-requirements": 0x00001000,
    "--retain-num-glyphs": 0x00002000,
    "--downgrade-cff2": 0x00004000,
}


class UserError(Exception):
    pass


def fail(msg: str) -> None:
    raise UserError(msg)


def pack_u32(value: int) -> bytes:
    return struct.pack("<I", value & 0xFFFFFFFF)


def pack_f32(value: float) -> bytes:
    return struct.pack("<f", value)


def emit_set_flags(ops: bytearray, flags: int) -> None:
    ops.append(HB_FUZZING_OP_SET_FLAGS)
    ops += pack_u32(flags)


def emit_keep_everything(ops: bytearray) -> None:
    ops.append(HB_FUZZING_OP_KEEP_EVERYTHING)


def emit_set_clear(ops: bytearray, set_type: int) -> None:
    ops += bytes((HB_FUZZING_OP_SET_CLEAR, set_type))


def emit_set_invert(ops: bytearray, set_type: int) -> None:
    ops += bytes((HB_FUZZING_OP_SET_INVERT, set_type))


def emit_set_ranges(ops: bytearray, opcode: int, set_type: int, ranges: list[tuple[int, int]]) -> None:
    ops += bytes((opcode, set_type))
    ops += pack_u32(len(ranges))
    for start, end in ranges:
        ops += pack_u32(start)
        ops += pack_u32(end)


def emit_text(ops: bytearray, opcode: int, codepoints: list[int]) -> None:
    ops.append(opcode)
    ops += pack_u32(len(codepoints))
    for cp in codepoints:
        ops += pack_u32(cp)


def emit_pin_all_axes_to_default(ops: bytearray) -> None:
    ops.append(HB_FUZZING_OP_AXIS_PIN_ALL_TO_DEFAULT)


def emit_axis_records(ops: bytearray, records: list[tuple[int, int, float, float, float]]) -> None:
    ops.append(HB_FUZZING_OP_AXIS_SET)
    ops += pack_u32(len(records))
    for tag, mode, minimum, middle, maximum in records:
        ops += pack_u32(tag)
        ops.append(mode)
        ops += pack_f32(minimum)
        ops += pack_f32(middle)
        ops += pack_f32(maximum)


def parse_decimal_ranges(value: str) -> list[tuple[int, int]]:
    ranges: list[tuple[int, int]] = []
    for chunk in value.replace(" ", ",").split(","):
        if not chunk:
            continue
        if chunk == "*":
            fail("Internal error: wildcard range should be handled before parsing.")
        if "-" in chunk:
            start_text, end_text = chunk.split("-", 1)
            try:
                start = int(start_text, 10)
                end = int(end_text, 10)
            except ValueError:
                fail(f"Failed parsing decimal range '{chunk}'.")
        else:
            try:
                start = end = int(chunk, 10)
            except ValueError:
                fail(f"Failed parsing decimal value '{chunk}'.")
        if start < 0 or end < start:
            fail(f"Invalid decimal range '{chunk}'.")
        ranges.append((start, end))
    return ranges


def parse_hex_ranges(value: str) -> list[tuple[int, int]]:
    ranges: list[tuple[int, int]] = []
    for chunk in value.replace(" ", ",").split(","):
        if not chunk:
            continue
        if chunk == "*":
            fail("Internal error: wildcard range should be handled before parsing.")
        if "-" in chunk:
            start_text, end_text = chunk.split("-", 1)
            try:
                start = int(start_text, 16)
                end = int(end_text, 16)
            except ValueError:
                fail(f"Failed parsing unicode range '{chunk}'.")
        else:
            try:
                start = end = int(chunk, 16)
            except ValueError:
                fail(f"Failed parsing unicode value '{chunk}'.")
        if start < 0 or end < start or end > 0x10FFFF:
            fail(f"Invalid unicode range '{chunk}'.")
        ranges.append((start, end))
    return ranges


def parse_tags(value: str) -> list[tuple[int, int]]:
    tags: list[tuple[int, int]] = []
    for chunk in value.replace(",", " ").split():
        if chunk == "*":
            fail("Internal error: wildcard tag set should be handled before parsing.")
        if len(chunk) > 4:
            fail(f"Tag '{chunk}' is longer than 4 bytes.")
        try:
            tag_value = int.from_bytes(chunk.ljust(4, " ").encode("ascii"), byteorder="big")
        except UnicodeEncodeError:
            fail(f"Tag '{chunk}' is not ASCII.")
        tags.append((tag_value, tag_value))
    return tags


def parse_text_codepoints(value: str) -> list[int]:
    codepoints = [ord(ch) for ch in value]
    for cp in codepoints:
        if cp < 0 or cp > 0x10FFFF:
            fail(f"Invalid codepoint value {cp}.")
    return codepoints


def parse_tag_value(tag_text: str, kind: str) -> int:
    if len(tag_text) > 4:
        fail(f"{kind} '{tag_text}' is longer than 4 bytes.")
    try:
        return int.from_bytes(tag_text.ljust(4, " ").encode("ascii"), byteorder="big")
    except UnicodeEncodeError:
        fail(f"{kind} '{tag_text}' is not ASCII.")


def parse_axis_number(text: str) -> float:
    try:
        return float(text)
    except ValueError:
        fail(f"Failed parsing axis value '{text}'.")


def parse_axis_range(value: str) -> tuple[int, float, float, float]:
    if value == "drop":
        return HB_FUZZING_AXIS_PIN_TO_DEFAULT, 0.0, 0.0, 0.0

    parts = value.split(":")
    if len(parts) == 1:
        v = parse_axis_number(parts[0])
        return HB_FUZZING_AXIS_SET_RANGE, v, v, v
    if len(parts) == 2:
        minimum = math.nan if parts[0] == "" else parse_axis_number(parts[0])
        maximum = math.nan if parts[1] == "" else parse_axis_number(parts[1])
        return HB_FUZZING_AXIS_SET_RANGE, minimum, math.nan, maximum
    if len(parts) == 3:
        minimum = math.nan if parts[0] == "" else parse_axis_number(parts[0])
        middle = math.nan if parts[1] == "" else parse_axis_number(parts[1])
        maximum = math.nan if parts[2] == "" else parse_axis_number(parts[2])
        return HB_FUZZING_AXIS_SET_RANGE, minimum, middle, maximum
    fail(f"Failed parsing axis range '{value}'.")


def parse_variations(value: str) -> tuple[bool, list[tuple[int, int, float, float, float]]]:
    pin_all = False
    records: list[tuple[int, int, float, float, float]] = []
    for item in value.replace(" ", ",").split(","):
        if not item:
            continue
        if "=" not in item:
            fail(f"Unsupported variation item '{item}'. Expected tag=value.")
        tag_text, axis_value_text = item.split("=", 1)
        if tag_text == "*":
            if axis_value_text != "drop":
                fail("Only *=drop is supported for wildcard axis pinning.")
            pin_all = True
            continue
        tag = parse_tag_value(tag_text, "Axis tag")
        mode, minimum, middle, maximum = parse_axis_range(axis_value_text)
        records.append((tag, mode, minimum, middle, maximum))
    return pin_all, records


def add_set_option(ops: bytearray, set_type: int, modifier: str, value: str, parser) -> None:
    if modifier == "":
        emit_set_clear(ops, set_type)

    if value == "*":
        if modifier == "-":
            emit_set_clear(ops, set_type)
        else:
            emit_set_clear(ops, set_type)
            emit_set_invert(ops, set_type)
        return

    ranges = parser(value)
    if ranges:
        emit_set_ranges(
            ops,
            HB_FUZZING_OP_SET_DEL_RANGES if modifier == "-" else HB_FUZZING_OP_SET_ADD_RANGES,
            set_type,
            ranges,
        )


def parse_output_arg(argv: list[str], usage_text: str) -> tuple[str, list[str]]:
    out = ""
    args = argv[1:]

    while args:
        arg = args[0]
        if arg == "-o":
            if len(args) < 2:
                print(usage_text, end="", file=sys.stderr)
                raise SystemExit(1)
            out = args[1]
            args = args[2:]
            continue
        if arg in ("-h", "--help"):
            print(usage_text, end="")
            raise SystemExit(0)
        if arg == "--":
            args = args[1:]
            break
        break

    return out, args


def read_text_file(path: str) -> list[int]:
    with open(path, "r", encoding="utf-8") as fp:
        return parse_text_codepoints(fp.read())


def read_range_file(path: str, parser) -> list[tuple[int, int]]:
    ranges = []
    with open(path, "r", encoding="utf-8") as fp:
        for line in fp:
            line = line.split("#", 1)[0].strip()
            if line:
                ranges.extend(parser(line))
    return ranges


def resolve_seed_output_path(font_path: str, font_bytes: bytes, out_path: str, default_suffix: str) -> tuple[str, str]:
    digest = hashlib.sha1(font_bytes).hexdigest()
    default_name = digest + default_suffix
    default_dir = os.path.join(os.path.dirname(__file__), "fonts")

    if not out_path:
        display_path = os.path.join(default_dir, default_name)
        return display_path, os.path.abspath(display_path)

    if out_path.endswith(os.sep):
        display_path = os.path.join(out_path, default_name)
        return display_path, os.path.abspath(display_path)

    abs_out_path = os.path.abspath(out_path)
    if os.path.isdir(abs_out_path):
        if os.path.isabs(out_path):
            display_path = os.path.join(out_path, default_name)
        else:
            display_path = os.path.join(out_path, default_name)
        return display_path, os.path.join(abs_out_path, default_name)

    return out_path, abs_out_path

def format_leading_comment(argv: list[str]) -> bytes:
    return ("$ " + " ".join(shlex.quote(arg) for arg in argv) + "\n").encode("utf-8")


def write_seed_file(font_file: str,
                    out_path: str,
                    ops: bytearray,
                    default_suffix: str = ".susbset-seed",
                    leading_comment: bytes = b"") -> str:
    font_path = os.path.abspath(font_file)
    if not os.path.isfile(font_path):
        fail(f"Font file not found: '{font_path}'.")

    with open(font_path, "rb") as fp:
        font_bytes = fp.read()

    display_path, seed_path = resolve_seed_output_path(font_path, font_bytes, out_path, default_suffix)
    parent = os.path.dirname(seed_path)
    if parent:
        os.makedirs(parent, exist_ok=True)

    with open(seed_path, "wb") as fp:
        fp.write(leading_comment)
        fp.write(font_bytes)
        fp.write(ops)
        fp.write(pack_u32(len(ops)))
        fp.write(HB_FUZZING_EXTENDED_MAGIC)

    return display_path


def print_recorded_seed_path(seed_path: str) -> None:
    print(seed_path)


def main_with_user_errors(main_fn, argv: list[str]) -> int:
    try:
        return main_fn(argv)
    except UserError as e:
        print(e, file=sys.stderr)
        return 1



#!/usr/bin/env python3

"This tool is intended to be used from meson"

import os, sys, shutil, pathlib

if len (sys.argv) < 3:
	sys.exit (__doc__)

OUTPUT = sys.argv[1]
CURRENT_SOURCE_DIR = sys.argv[2]

# make sure input files are unique
sources = sorted(set(sys.argv[3:]))

with open (OUTPUT, "wb") as f:
	f.write ("".join ('#include "{}"\n'.format (pathlib.Path( os.path.relpath (os.path.abspath (x), CURRENT_SOURCE_DIR) ).as_posix()) for x in sources if x.endswith (".cc")).encode ())

# copy it also to the source tree, but only if it has changed
baseline_filename = os.path.join (CURRENT_SOURCE_DIR, os.path.basename (OUTPUT))
with open(baseline_filename, "rb") as baseline:
	with open(OUTPUT, "rb") as generated:
		if baseline.read() != generated.read():
			shutil.copyfile (OUTPUT, baseline_filename)



#!/usr/bin/env python3

import re
import argparse

parser = argparse.ArgumentParser ()
parser.add_argument ('input')
parser.add_argument ('output')
args = parser.parse_args ()

with open (args.input, 'r') as inp, open (args.output, 'w') as out:
    for l in inp.readlines ():
        l = re.sub ('_t_get_type', '_get_type', l)
        l = re.sub (r'_T \(', ' (', l)
        out.write (l)



#!/usr/bin/env python3

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, HarfBuzz as hb


POOL = {}


def move_to_f(funcs, draw_data, st, to_x, to_y, user_data):
    context = POOL[draw_data]
    context.move_to(to_x, to_y)


def line_to_f(funcs, draw_data, st, to_x, to_y, user_data):
    context = POOL[draw_data]
    context.line_to(to_x, to_y)


def cubic_to_f(
    funcs,
    draw_data,
    st,
    control1_x,
    control1_y,
    control2_x,
    control2_y,
    to_x,
    to_y,
    user_data,
):
    context = POOL[draw_data]
    context.curve_to(control1_x, control1_y, control2_x, control2_y, to_x, to_y)


def close_path_f(funcs, draw_data, st, user_data):
    context = POOL[draw_data]
    context.close_path()


DFUNCS = hb.draw_funcs_create()
hb.draw_funcs_set_move_to_func(DFUNCS, move_to_f, None)
hb.draw_funcs_set_line_to_func(DFUNCS, line_to_f, None)
hb.draw_funcs_set_cubic_to_func(DFUNCS, cubic_to_f, None)
hb.draw_funcs_set_close_path_func(DFUNCS, close_path_f, None)


def push_transform_f(funcs, paint_data, xx, yx, xy, yy, dx, dy, user_data):
    raise NotImplementedError


def pop_transform_f(funcs, paint_data, user_data):
    raise NotImplementedError


def color_f(funcs, paint_data, is_foreground, color, user_data):
    context = POOL[paint_data]
    r = hb.color_get_red(color) / 255
    g = hb.color_get_green(color) / 255
    b = hb.color_get_blue(color) / 255
    a = hb.color_get_alpha(color) / 255
    context.set_source_rgba(r, g, b, a)
    context.paint()


def push_clip_rectangle_f(funcs, paint_data, xmin, ymin, xmax, ymax, user_data):
    context = POOL[paint_data]
    context.save()
    context.rectangle(xmin, ymin, xmax, ymax)
    context.clip()


def push_clip_glyph_f(funcs, paint_data, glyph, font, user_data):
    context = POOL[paint_data]
    context.save()
    context.new_path()
    hb.font_draw_glyph(font, glyph, DFUNCS, paint_data)
    context.close_path()
    context.clip()


def pop_clip_f(funcs, paint_data, user_data):
    context = POOL[paint_data]
    context.restore()


def push_group_f(funcs, paint_data, user_data):
    raise NotImplementedError


def pop_group_f(funcs, paint_data, mode, user_data):
    raise NotImplementedError


PFUNCS = hb.paint_funcs_create()
hb.paint_funcs_set_push_transform_func(PFUNCS, push_transform_f, None)
hb.paint_funcs_set_pop_transform_func(PFUNCS, pop_transform_f, None)
hb.paint_funcs_set_color_func(PFUNCS, color_f, None)
hb.paint_funcs_set_push_clip_glyph_func(PFUNCS, push_clip_glyph_f, None)
hb.paint_funcs_set_push_clip_rectangle_func(PFUNCS, push_clip_rectangle_f, None)
hb.paint_funcs_set_pop_clip_func(PFUNCS, pop_clip_f, None)
hb.paint_funcs_set_push_group_func(PFUNCS, push_group_f, None)
hb.paint_funcs_set_pop_group_func(PFUNCS, pop_group_f, None)


def makebuffer(words):
    buf = hb.buffer_create()

    text = " ".join(words)
    hb.buffer_add_codepoints(buf, [ord(c) for c in text], 0, len(text))

    hb.buffer_guess_segment_properties(buf)

    return buf


def justify(face, words, advance, target_advance):
    font = hb.font_create(face)
    buf = makebuffer(words)

    wiggle = 5
    shrink = target_advance - wiggle < advance
    expand = target_advance + wiggle > advance

    ret, advance, tag, value = hb.shape_justify(
        font,
        buf,
        None,
        None,
        target_advance,
        target_advance,
        advance,
    )

    if not ret:
        return False, buf, None

    if tag:
        variation = hb.variation_t()
        variation.tag = tag
        variation.value = value
    else:
        variation = None

    if shrink and advance > target_advance + wiggle:
        return False, buf, variation
    if expand and advance < target_advance - wiggle:
        return False, buf, variation

    return True, buf, variation


def shape(face, words):
    font = hb.font_create(face)
    buf = makebuffer(words)
    hb.shape(font, buf)
    positions = hb.buffer_get_glyph_positions(buf)
    advance = sum(p.x_advance for p in positions)
    return buf, advance


def typeset(face, text, target_advance):
    lines = []
    words = []
    for word in text.split():
        words.append(word)
        buf, advance = shape(face, words)
        if advance > target_advance:
            # Shrink
            ret, buf, variation = justify(face, words, advance, target_advance)
            if ret:
                lines.append((buf, variation))
                words = []
            # If if fails, pop the last word and shrink, and hope for the best.
            # A too short line is better than too long.
            elif len(words) > 1:
                words.pop()
                _, buf, variation = justify(face, words, advance, target_advance)
                lines.append((buf, variation))
                words = [word]
            # But if it is one word, meh.
            else:
                lines.append((buf, variation))
                words = []

    # Justify last line
    if words:
        _, buf, variation = justify(face, words, advance, target_advance)
        lines.append((buf, variation))

    return lines


def render(face, text, context, width, height, fontsize):
    font = hb.font_create(face)

    margin = fontsize * 2
    scale = fontsize / hb.face_get_upem(face)
    target_advance = (width - (margin * 2)) / scale

    lines = typeset(face, text, target_advance)

    _, extents = hb.font_get_h_extents(font)
    lineheight = extents.ascender - extents.descender + extents.line_gap
    lineheight *= scale

    context.save()
    context.translate(0, margin)
    context.set_font_size(12)
    context.set_source_rgb(1, 0, 0)
    for buf, variation in lines:
        rtl = hb.buffer_get_direction(buf) == hb.direction_t.RTL
        if rtl:
            hb.buffer_reverse(buf)
        infos = hb.buffer_get_glyph_infos(buf)
        positions = hb.buffer_get_glyph_positions(buf)
        advance = sum(p.x_advance for p in positions)

        context.translate(0, lineheight)
        context.save()

        context.save()
        context.move_to(0, -20)
        if variation:
            tag = hb.tag_to_string(variation.tag).decode("ascii")
            context.show_text(f" {tag}={variation.value:g}")
        context.move_to(0, 0)
        context.show_text(f" {advance:g}/{target_advance:g}")
        context.restore()

        if variation:
            hb.font_set_variations(font, [variation])

        context.translate(margin, 0)
        context.scale(scale, -scale)

        if rtl:
            context.translate(target_advance, 0)

        for info, pos in zip(infos, positions):
            if rtl:
                context.translate(-pos.x_advance, pos.y_advance)
            context.save()
            context.translate(pos.x_offset, pos.y_offset)
            hb.font_paint_glyph(font, info.codepoint, PFUNCS, id(context), 0, 0x0000FF)
            context.restore()
            if not rtl:
                context.translate(+pos.x_advance, pos.y_advance)

        context.restore()
    context.restore()


def main(fontpath, textpath):
    fontsize = 70

    blob = hb.blob_create_from_file(fontpath)
    face = hb.face_create(blob, 0)

    with open(textpath) as f:
        text = f.read()

    def on_draw(da, context):
        alloc = da.get_allocation()
        POOL[id(context)] = context
        render(face, text, context, alloc.width, alloc.height, fontsize)
        del POOL[id(context)]

    drawingarea = Gtk.DrawingArea()
    drawingarea.connect("draw", on_draw)

    win = Gtk.Window()
    win.connect("destroy", Gtk.main_quit)
    win.set_default_size(1000, 700)
    win.add(drawingarea)

    win.show_all()
    Gtk.main()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="HarfBuzz justification demo.")
    parser.add_argument("fontfile", help="font file")
    parser.add_argument("textfile", help="text")
    args = parser.parse_args()
    main(args.fontfile, args.textfile)
